// UC: 21020 - Computação Gráfica (2022/23)
// Aluno: 1902551 - Pedro Gaspar

import * as THREE from "https://unpkg.com/three@0.124.0/build/three.module.js";
import { OrbitControls } from "https://unpkg.com/three@0.124.0/examples/jsm/controls/OrbitControls.js";
import { lineMP } from "../lineMP.mjs";

// /////////////////////////////////////////////////////////////////////////////
// #region Variáveis globais

let screenWidth = window.innerWidth;
let screenHeight = window.innerHeight;

const TILE_WIDTH = 40;
const TILES_PER_ROW = 21;
const GRID_WIDTH = TILES_PER_ROW * TILE_WIDTH;

const CAMERA = new THREE.PerspectiveCamera(
  45, // fov
  screenWidth / screenHeight, // aspect
  1, // near
  10000 // far
);

const SCENE = new THREE.Scene();

const RENDERER = new THREE.WebGLRenderer({ antialias: true });

const PLANE = new THREE.Mesh(
  new THREE.PlaneGeometry(GRID_WIDTH, GRID_WIDTH),
  new THREE.MeshBasicMaterial({ visible: false })
);
const CONTROLS = new OrbitControls(CAMERA, RENDERER.domElement);

const HOVER_TILE = new THREE.Mesh(
  new THREE.BoxGeometry(TILE_WIDTH, 0, TILE_WIDTH),
  new THREE.MeshBasicMaterial({
    color: "#f00",
    opacity: 0.25,
    transparent: true,
  })
);

const SCENE_OBJECTS = [];

/** Primeiro ladrilho onde o utilizador clicou para desenhar a linha */
let startTile;

// #endregion
// /////////////////////////////////////////////////////////////////////////////

/** Inicialização. Renderiza o plano e define os event listeners. */
function init() {
  CAMERA.position.set(-75, 1000, 800);
  CAMERA.lookAt(0, 0, 0);

  SCENE.background = new THREE.Color("#fff");

  HOVER_TILE.visible = false;
  SCENE.add(HOVER_TILE);

  PLANE.rotateX(-(Math.PI / 2));
  SCENE.add(PLANE);

  RENDERER.setSize(screenWidth, screenHeight);
  RENDERER.setPixelRatio(window.devicePixelRatio);
  document.body.appendChild(RENDERER.domElement);

  // prevenir a câmara de rodar por baixo do plano
  CONTROLS.maxPolarAngle = Math.PI / 2;
  CONTROLS.update();

  drawGrid();

  const raycaster = new THREE.Raycaster();
  const pointer = new THREE.Vector2();

  // eventos
  document.addEventListener("pointermove", (event) => {
    const x = (event.clientX / screenWidth) * 2 - 1;
    const y = -(event.clientY / screenHeight) * 2 + 1;

    pointer.set(x, y);
    raycaster.setFromCamera(pointer, CAMERA);

    const intersections = raycaster.intersectObjects(
      SCENE.children.filter((child) => child.name === "tile"),
      false
    );

    if (intersections.length) {
      const [intersection] = intersections;

      if (intersection.face.normal) {
        const { x, z } = intersection.object.position;

        if (HOVER_TILE.position.x !== x || HOVER_TILE.position.z !== z) {
          // Faz log para a consola quando o ladrilho muda, como dita o enunciado
          console.log(`(${x / TILE_WIDTH}, ${z / TILE_WIDTH})`);

          // Coloca a posição do ladrilho de hover no ladrilho onde o rato está
          HOVER_TILE.position.x = x;
          HOVER_TILE.position.z = z;
          HOVER_TILE.visible = true;
        }
      }
    }

    render();
  });

  document.addEventListener("keydown", (event) => {
    switch (event.code) {
      case "Backspace":
        clearScene();
        break;
      case "KeyX":
        onTileSelect();
        break;
    }
  });

  window.addEventListener("resize", () => {
    screenWidth = window.innerWidth;
    screenHeight = window.innerHeight;

    CAMERA.aspect = screenWidth / screenHeight;
    CAMERA.updateProjectionMatrix();

    RENDERER.setSize(screenWidth, screenHeight);

    render();
  });
}

/**
 * Desenha os ladrilhos na grelha, assim como a linhas das abcissas e ordenadas
 */
function drawGrid() {
  const grid = new THREE.PlaneGeometry(TILE_WIDTH, TILE_WIDTH);

  const tileIndices = [...Array(TILES_PER_ROW).keys()].map(
    (x) => (x -= Math.floor(TILES_PER_ROW / 2))
  );
  for (const row of tileIndices) {
    for (const col of tileIndices) {
      const tileColor = (row + col) % 2 === 0 ? "#f98866" : "#9089b3";

      const tile = new THREE.Mesh(
        grid,
        new THREE.LineBasicMaterial({
          color: tileColor,
          linewidth: 10,
        })
      );

      // O nome é utilizado para detectar a interseção do rato com os ladrilhos
      tile.name = "tile";
      tile.position.x = row * TILE_WIDTH;
      tile.position.z = col * TILE_WIDTH;
      tile.position.y = 0;
      // Rodar 90 graus
      tile.rotateX(-Math.PI / 2);
      SCENE.add(tile);
    }
  }

  // Linha das ordenadas
  const yLineCoords = [
    new THREE.Vector3(0, 0, 0),
    new THREE.Vector3(0, 0, -GRID_WIDTH / 2),
  ];

  const yLine = new THREE.Line(
    new THREE.BufferGeometry().setFromPoints(yLineCoords),
    new THREE.LineBasicMaterial({ color: "#f00" })
  );

  SCENE.add(yLine);

  // Linha das abcissas
  const xLineCoords = [
    new THREE.Vector3(0, 0, 0),
    new THREE.Vector3(GRID_WIDTH / 2, 0, 0),
  ];
  const xLine = new THREE.Line(
    new THREE.BufferGeometry().setFromPoints(xLineCoords),
    new THREE.LineBasicMaterial({ color: "#00f" })
  );

  SCENE.add(xLine);
}

/**
 * Controla a seleção de um ladrilho. Se selecionou o primeiro ladrilho,
 * guarda-o na variável global startTile. Se selecionou o segundo ladrilho,
 * desenha a linha entre os dois ladrilhos e preenche os ladrilhos
 * correspondentes ao ponto médio entre eles.
 */
function onTileSelect() {
  const tile = new THREE.Mesh(
    new THREE.BoxGeometry(TILE_WIDTH, 0, TILE_WIDTH),
    new THREE.MeshBasicMaterial({ color: "#f00" })
  );
  tile.position.add(HOVER_TILE.position);
  SCENE.add(tile);
  SCENE_OBJECTS.push(tile);

  render();

  if (startTile == null) {
    // O utilizador acabou de seleccionar o ladrilho de início
    startTile = tile;
    return;
  }

  if (
    tile.position.x === startTile.position.x &&
    tile.position.z === startTile.position.z
  ) {
    // Se seleccionou o mesmo ladrilho, não faz nada
    alert("Selecione um ladrilho diferente do primeiro.");
    return;
  }

  // O utilizador acabou de seleccionar o ladrilho de fim.
  // Vamos desenhar a linha e preencher os ladrilhos do ponto médio

  fillMidpointTiles(startTile, tile);
  drawSelectionLine(startTile, tile);

  startTile = undefined;

  render();
}

/**
 * Preenche os ladrilhos correspondentes ao ponto médio entre dois ladrilhos
 * @param {THREE.Object3D} startTile Ladrilho de início
 * @param {THREE.Object3D} endTile Ladrilho de fim
 */
function fillMidpointTiles(startTile, endTile) {
  const startCoords = {
    x: startTile.position.x / TILE_WIDTH,
    y: startTile.position.z / TILE_WIDTH,
  };

  const endCoords = {
    x: endTile.position.x / TILE_WIDTH,
    y: endTile.position.z / TILE_WIDTH,
  };

  const midpointTiles = lineMP(startCoords, endCoords);
  midpointTiles.forEach(({ x, y }) => {
    const selectionTile = new THREE.Mesh(
      new THREE.BoxGeometry(TILE_WIDTH, TILE_WIDTH * 0.25, TILE_WIDTH),
      new THREE.MeshBasicMaterial({
        color: "#ff0",
        opacity: 0.5,
        transparent: true,
      })
    );

    selectionTile.position.add({
      x: (x *= TILE_WIDTH),
      y: TILE_WIDTH / 8,
      z: (y *= TILE_WIDTH),
    });

    SCENE.add(selectionTile);
    SCENE_OBJECTS.push(selectionTile);
  });
}

/**
 * Desenha a linha reta entre o centro do ladrilho de início e o centro
 * do ladrilho de fim
 * @param {THREE.Object3D} startTile Ladrilho de início
 * @param {THREE.Object3D} endTile Ladrilho de fim
 */
function drawSelectionLine(startTile, endTile) {
  const lineGeo = new THREE.BufferGeometry().setFromPoints([
    new THREE.Vector3(startTile.position.x, 0, startTile.position.z),
    new THREE.Vector3(endTile.position.x, 0, endTile.position.z),
  ]);

  const line = new THREE.Line(
    lineGeo,
    new THREE.LineBasicMaterial({ color: "#000" })
  );

  SCENE.add(line);
  SCENE_OBJECTS.push(line);
}

/**
 * Limpar todos os objetos introduzidos na cena pelo utilizador
 */
function clearScene() {
  let obj;

  while ((obj = SCENE_OBJECTS.pop())) {
    SCENE.remove(obj);
  }

  render();
}

function render() {
  RENDERER.render(SCENE, CAMERA);
}

// /////////////////////////////////////////////////////////////////////////////

init();
render();
